/////////////////////////////////////////////////////////////////////////
// $Id: instrument.cc 12655 2015-02-19 20:23:08Z sshwarts $
/////////////////////////////////////////////////////////////////////////
//
//   Copyright (c) 2006-2015 Stanislav Shwartsman
//          Written by Stanislav Shwartsman [sshwarts at sourceforge net]
//
//  This library is free software; you can redistribute it and/or
//  modify it under the terms of the GNU Lesser General Public
//  License as published by the Free Software Foundation; either
//  version 2 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
//  Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public
//  License along with this library; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA


#include "bochs.h"
#define NEED_CPU_REG_SHORTCUTS 1
#include "cpu/cpu.h"
#include "disasm/disasm.h"
#include <cstdio>

#if BX_INSTRUMENTATION

FILE *logfile;
int logmax;
static int logstart = 4000000;

void bx_instr_init_env(void)
{
	logfile = NULL;
	logmax = 0;
}
void bx_instr_exit_env(void)
{
	if(logfile) {
		fclose(logfile);
	}
}

void bx_instr_initialize(unsigned cpu)
{
	if(logfile) {
		fclose(logfile);
	}
	logfile = NULL;
	logmax = 0;
	if(SIM->get_param_bool("ibmulator.logging")->get()) {
		logfile = fopen(SIM->get_param_string("ibmulator.file")->getptr(), "w");
		if(!logfile) {
			fprintf(stderr, "unable to open file %s\n", SIM->get_param_string("ibmulator.file")->getptr());
		} else {
			logmax = SIM->get_param_num("ibmulator.count")->get();
			fprintf(stderr, "logging %d cpu instructions to file %s\n", logmax,
					SIM->get_param_string("ibmulator.file")->getptr());
		}
	}
}
void bx_instr_exit(unsigned cpu) {}
void bx_instr_reset(unsigned cpu, unsigned type) {}
void bx_instr_hlt(unsigned cpu) {}
void bx_instr_mwait(unsigned cpu, bx_phy_address addr, unsigned len, Bit32u flags) {}

void bx_instr_debug_promt() {}
void bx_instr_debug_cmd(const char *cmd) {}

void bx_instr_cnear_branch_taken(unsigned cpu, bx_address branch_eip, bx_address new_eip) {}
void bx_instr_cnear_branch_not_taken(unsigned cpu, bx_address branch_eip) {}
void bx_instr_ucnear_branch(unsigned cpu, unsigned what, bx_address branch_eip, bx_address new_eip) {}
void bx_instr_far_branch(unsigned cpu, unsigned what, Bit16u prev_cs, bx_address prev_eip, Bit16u new_cs, bx_address new_eip) {}

void bx_instr_opcode(unsigned cpu, bxInstruction_c *i, const Bit8u *opcode, unsigned len, bx_bool is32, bx_bool is64) {}

void bx_instr_interrupt(unsigned cpu, unsigned vector) {}
void bx_instr_exception(unsigned cpu, unsigned vector, unsigned error_code) {}
void bx_instr_hwinterrupt(unsigned cpu, unsigned vector, Bit16u cs, bx_address eip) {}

void bx_instr_tlb_cntrl(unsigned cpu, unsigned what, bx_phy_address new_cr3) {}
void bx_instr_clflush(unsigned cpu, bx_address laddr, bx_phy_address paddr) {}
void bx_instr_cache_cntrl(unsigned cpu, unsigned what) {}
void bx_instr_prefetch_hint(unsigned cpu, unsigned what, unsigned seg, bx_address offset) {}

#define PCPULOG(_dest_, _format_, ...) { \
	if(fprintf(_dest_, _format_, ## __VA_ARGS__) < 0) \
		return; \
}

void write_segreg(FILE *_dest, unsigned cpu, unsigned _segreg, const char *_name)
{
	PCPULOG(_dest, "%s=[%04X", _name, BX_CPU(cpu)->sregs[_segreg].selector.value);

	PCPULOG(_dest, " %s ", BX_CPU(cpu)->sregs[_segreg].cache.segment?"S":"s");
	if(BX_CPU(cpu)->real_mode() || BX_CPU(cpu)->sregs[_segreg].cache.segment) {
		PCPULOG(_dest, "%08X-%08X",
				BX_CPU(cpu)->sregs[_segreg].cache.u.segment.base,
				BX_CPU(cpu)->sregs[_segreg].cache.u.segment.limit_scaled);
	}
	PCPULOG(_dest, " %02X ", get_ar_byte(&BX_CPU(cpu)->sregs[_segreg].cache));
	if(BX_CPU(cpu)->real_mode() || BX_CPU(cpu)->sregs[_segreg].cache.segment) {
		PCPULOG(_dest, "%s%s",
			BX_CPU(cpu)->sregs[_segreg].cache.u.segment.d_b?"B":"b",
			BX_CPU(cpu)->sregs[_segreg].cache.u.segment.g?"G":"g");

	}
	PCPULOG(_dest, "%s] ",
		(BX_CPU(cpu)->sregs[_segreg].cache.valid&SegValidCache) ? "V":"v");
}

void bx_instr_before_execution(unsigned cpu, bxInstruction_c *_i)
{
	static bx_address lastIP = 0;
	static bx_address lastCS = 0;
	static int logcount = 0;

	if(_i->repUsedL() && lastIP==EIP && lastCS==BX_CPU(cpu)->sregs[BX_SEG_REG_CS].selector.value) {
		// don't log repetitions of the same instruction
		return;
	}

	lastIP = EIP;
	lastCS = BX_CPU(cpu)->sregs[BX_SEG_REG_CS].selector.value;

	logcount++;
	if(logmax==0 || logstart>=logcount) {
		return;
	}
	logmax--;

	if(fprintf(logfile, "%04X:%08X ", lastCS, lastIP) < 0)
		return;

	for(unsigned j=0; j<15; j++) {
		if(j < _i->ilen()) {
			if(fprintf(logfile, "%02X ", _i->opcode_bytes[j]) < 0)
				return;
		} else {
			if(fprintf(logfile, "   ") < 0)
				return;
		}
	}

	/*
	if(fprintf(logfile, "SE=%d,SM=%d,SA=%d ",
			BX_CPU(cpu)->pending_event,
			BX_CPU(cpu)->event_mask,
			BX_CPU(cpu)->async_event
			) < 0)
		return;
	*/

	if(fprintf(logfile, "EF=%05X ", BX_CPU(cpu)->read_eflags()) < 0)
		return;
	if(fprintf(logfile, "EAX=%08X EBX=%08X ECX=%08X EDX=%08X ", EAX, EBX, ECX, EDX) < 0)
		return;
	if(fprintf(logfile, "ESI=%08X EDI=%08X EBP=%08X ESP=%08X ", ESI, EDI, EBP, ESP) < 0)
		return;

	write_segreg(logfile, cpu, BX_SEG_REG_CS, "CS");
	write_segreg(logfile, cpu, BX_SEG_REG_ES, "ES");
	write_segreg(logfile, cpu, BX_SEG_REG_DS, "DS");
	write_segreg(logfile, cpu, BX_SEG_REG_SS, "SS");
	write_segreg(logfile, cpu, BX_SEG_REG_FS, "FS");
	write_segreg(logfile, cpu, BX_SEG_REG_GS, "GS");

	bx_address cr0 = BX_CPU(cpu)->read_CR0();
	if(fprintf(logfile, "CR0=PE:%d,TS:%d,PG:%d ",
			bool(cr0 & 0x1),
			bool(cr0 & 0x8),
			bool(cr0 & 0x80000000)) < 0)
		return;
	if(fprintf(logfile, "CR2=%08X CR3=%08X ",
			BX_CPU(cpu)->cr2,
			BX_CPU(cpu)->cr3) < 0)
		return;

	if(fprintf(logfile, "\n") < 0)
		return;
}

void bx_instr_after_execution(unsigned cpu, bxInstruction_c *i) {}
void bx_instr_repeat_iteration(unsigned cpu, bxInstruction_c *i) {}

void bx_instr_inp(Bit16u addr, unsigned len) {}
void bx_instr_inp2(Bit16u addr, unsigned len, unsigned val) {}
void bx_instr_outp(Bit16u addr, unsigned len, unsigned val) {}

void bx_instr_lin_access(unsigned cpu, bx_address lin, bx_address phy, unsigned len, unsigned memtype, unsigned rw) {}
void bx_instr_phy_access(unsigned cpu,                 bx_address phy, unsigned len, unsigned memtype, unsigned rw) {}

void bx_instr_wrmsr(unsigned cpu, unsigned addr, Bit64u value) {}

void bx_instr_vmexit(unsigned cpu, Bit32u reason, Bit64u qualification) {}

#endif
